import { world, GameMode, ItemLockMode, ItemDurabilityComponent, ItemEnchantableComponent, system, ItemStack, EntityEquippableComponent, EquipmentSlot } from '@minecraft/server';

/**
 * คำนวณจำนวนเกมติ๊กตามระยะเวลาที่กำหนด
 * @param {number} [milliseconds=1000] - ระยะเวลาในมิลลิวินาที (ค่าเริ่มต้นคือ 1000)
 * @param {number} [gameTicksPerSecond=20] - จำนวนเกมติ๊กต่อวินาที (ค่าเริ่มต้นคือ 20)
 * @param {number} [millisecondsPerSecond=1000] - จำนวนมิลลิวินาทีต่อวินาที (ค่าเริ่มต้นคือ 1000)
 * @returns {number} จำนวนเกมติ๊กที่คำนวณได้
 * @example
 * ```js
 * calcGameTicks(1000) // => 20
 * calcGameTicks(500) // => 10
 * ```
 */
function calcGameTicks(milliseconds = 1000, gameTicksPerSecond = 20, millisecondsPerSecond = 1000) {
    return Math.floor(milliseconds / (millisecondsPerSecond / gameTicksPerSecond));
}

/**
 * จาก {@link location} เป็นจุดศูนย์กลาง ให้ได้พิกัดช่วงของรัศมี {@link radius}
 * @param { Location } location
 * @param { number } [radius=1]
 * @returns
 */
function getRadiusRange(location, radius = 1) {
    const centerX = location.x;
    const centerY = location.y;
    const centerZ = location.z;
    /*
        เก็บลิสต์ของอ็อบเจ็กต์สี่เหลี่ยมขนาด 3x3 โดยมีพิกัดของจัตุรัสปัจจุบันเป็นศูนย์กลาง
  
        มุมมองจากด้านบน: 0 คือจัตุรัสปัจจุบัน ได้พิกัดของทั้งหมดที่เป็น 1
  
        ชั้นแรก
        111
        111
        111
  
        ชั้นที่สอง
        111
        101
        111
  
        ชั้นที่สาม
        111
        111
        111
        */
    const positions = [];
    for (let x = centerX - radius; x <= centerX + radius; x++) {
        for (let y = centerY - radius; y <= centerY + radius; y++) {
            for (let z = centerZ - radius; z <= centerZ + radius; z++) {
                positions.push({ x, y, z });
            }
        }
    }
    return positions;
}

/**
 * แบ่งจำนวนไอเท็มออกเป็นกลุ่ม โดยแต่ละกลุ่มมีจำนวนเท่าไหร่
 * @param {number} sum จำนวนรวมทั้งหมด
 * @param {number} groupSize ขนาดต่อกลุ่ม (ค่าเริ่มต้น 64)
 * @returns {number[]}
 * @example
 * ```js
 * splitGroups(65) // => [64, 1]
 * splitGroups(140) // => [64, 64, 12]
 * splitGroups(65, 16) // => [16, 16, 16, 16, 1]
 * ```
 */
function splitGroups(sum, groupSize = 64) {
    const groups = [];
    while (sum > 0) {
        const group = Math.min(sum, groupSize);
        groups.push(group);
        sum -= group;
    }
    return groups;
}

function isSurvivalPlayer(dimension, player) {
    return dimension.getPlayers({ gameMode: GameMode.survival }).some((p) => p.name === player.name);
}
const isStrippedLog = (typeId) => typeId.includes('stripped_');
const getPlayerMainhand = (player) => {
    const entityEquippableComponent = player.getComponent(EntityEquippableComponent.componentId);
    return entityEquippableComponent.getEquipmentSlot(EquipmentSlot.Mainhand);
};
// ผู้เล่นไม่ได้ลอบเดินหรือไม่ได้ถือขวาน หนึ่งในเงื่อนไขไม่ตรงก็จะจบการทำงานโดยตรง
const getPlayerAction = (player) => player.isSneaking && getPlayerMainhand(player)?.hasTag('is_axe');
function isTree(dimension, locations) {
    const leaves = ['leaves', 'warped_wart_block', 'nether_wart_block'];
    for (const location of locations) {
        const blocksLocation = getRadiusRange(location);
        const is = blocksLocation.some((block) => {
            const typeId = dimension.getBlock(block)?.typeId;
            if (!typeId)
                return false;
            return leaves.some((item) => typeId.includes(item));
        });
        if (is)
            return true;
    }
    return false;
}
function consumeAxeDurability(player, logLocations) {
    const mainHand = getPlayerMainhand(player);
    try {
        if (!mainHand)
            return;
        mainHand.lockMode = ItemLockMode.slot;
        const item = mainHand.getItem();
        if (!item)
            return;
        const itemDurability = item.getComponent(ItemDurabilityComponent.componentId);
        const enchantments = item.getComponent(ItemEnchantableComponent.componentId);
        if (!enchantments || !itemDurability)
            return;
        const unbreaking = enchantments.getEnchantment('unbreaking')?.level || 0;
        // https://minecraft.fandom.com/wiki/Unbreaking
        const itemMaxDamage = itemDurability.damage * (1 + unbreaking);
        const itemMaxDurability = itemDurability.maxDurability * (1 + unbreaking);
        const consumeItemMaxDamage = itemMaxDamage + logLocations.length;
        const overproof = consumeItemMaxDamage >= itemMaxDurability ? consumeItemMaxDamage - itemMaxDurability : 0;
        if (overproof > 0)
            logLocations.splice(-overproof);
        // กำหนดระดับความเสียหายของขวาน
        const damage = Math.ceil((consumeItemMaxDamage * 1) / (1 + unbreaking));
        itemDurability.damage = damage > itemDurability.maxDurability ? itemDurability.maxDurability : damage;
        mainHand.setItem(item);
    }
    catch (error) {
        // eslint-disable-next-line no-console
        console.warn('เกิดข้อผิดพลาดในการตัดต้นไม้:', error);
    }
    finally {
        system.runTimeout(() => {
            if (mainHand)
                mainHand.lockMode = ItemLockMode.none;
        }, calcGameTicks(1000));
    }
}
function getLogLocations(dimension, location, currentBreakBlockTypeId) {
    const visited = new Set();
    const logLocations = [];
    const locations = getRadiusRange(location);
    const reg = /(_log|crimson_stem|warped_stem)$/;
    for (const vector3 of locations) {
        const pos = Object.values(vector3).join(',');
        if (visited.has(pos))
            continue;
        visited.add(pos);
        const block = dimension.getBlock(vector3);
        if (block && block.typeId === currentBreakBlockTypeId && reg.test(block.typeId)) {
            logLocations.push(vector3);
            locations.push(...getRadiusRange(block.location));
        }
    }
    return logLocations.sort((a, b) => a.y - b.y);
}
async function treeCut(location, dimension, logLocations) {
    if (!logLocations.length)
        return;
    const block = dimension.getBlock(logLocations[0]);
    const typeId = block.typeId;
    for (const location of logLocations) {
        const block = dimension.getBlock(location);
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        block.setType('air');
    }
    splitGroups(logLocations.length).forEach((group) => {
        dimension.spawnItem(new ItemStack(typeId, group), location);
    });
}
// eslint-disable-next-line max-statements
async function clearLeaves(dimension, logLocations) {
    const visited = new Set();
    const batchSize = 27; // ขนาดของแต่ละชุด
    let counter = 0;
    for (const logLocation of logLocations) {
        const locations = getRadiusRange(logLocation);
        for (const location of locations) {
            const pos = Object.values(location).join(',');
            if (visited.has(pos))
                continue;
            visited.add(pos);
            const block = dimension.getBlock(location);
            if (block && block.typeId.includes('leaves')) {
                const isIncludesLog = getRadiusRange(block.location, 2).some((location) => {
                    const block = dimension.getBlock(location);
                    if (!block)
                        return false;
                    const typeId = block.typeId;
                    if (isStrippedLog(typeId))
                        return false;
                    if (/_log$/.test(typeId))
                        return true;
                });
                // eslint-disable-next-line max-depth
                if (isIncludesLog)
                    continue;
                locations.push(...getRadiusRange(block.location));
                // eslint-disable-next-line max-depth
                if (counter === batchSize) {
                    // เพิ่มการหน่วงเวลาสั้น ๆ เพื่อให้ event loop ทำงาน
                    await new Promise((resolve) => system.runTimeout(resolve));
                    counter = 0;
                }
                counter++;
                const command = `setblock ${Object.values(location).join(' ')} air destroy`;
                dimension.runCommandAsync(command);
            }
        }
    }
}
world.afterEvents.playerBreakBlock.subscribe(async (e) => {
    try {
        const { dimension, player, block } = e;
        const currentBreakBlock = e.brokenBlockPermutation;
        const currentBreakBlockTypeId = currentBreakBlock.type.id;
        if (isStrippedLog(currentBreakBlockTypeId))
            return;
        const action = getPlayerAction(player);
        if (!action)
            return;
        const logLocations = getLogLocations(dimension, block.location, currentBreakBlockTypeId);
        const _isTree = isTree(dimension, logLocations);
        if (!_isTree)
            return;
        const survivalPlayer = isSurvivalPlayer(dimension, player);
        if (survivalPlayer)
            consumeAxeDurability(player, logLocations);
        await treeCut(block.location, dimension, logLocations);
        clearLeaves(dimension, logLocations);
    }
    catch (error) {
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const err = error;
        /* eslint-disable no-console */
        console.log('ข้อผิดพลาด', err);
        console.log('error.stack', err && err.stack);
        console.log('error.message', err && err.message);
        /* eslint-enable no-console */
    }
});